@@ -1,5 +1,5 @@
 # VMware vSphere Python SDK
-# Copyright (c) 2008-2014 VMware, Inc. All Rights Reserved.
+# Copyright (c) 2008-2018 VMware, Inc. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,17 +21,16 @@ from six import iteritems
 from six import iterkeys
 from six import itervalues
 from six import text_type
+from six import string_types
+from six import binary_type
 from six import PY3
 from datetime import datetime
-import pyVmomi.Iso8601
+from pyVmomi import Iso8601
 import base64
+import json
 import threading
-
 if PY3:
-    # python3 removed long, it's the same as int
-    long = int
-    # python3 removed basestring, use str instead.
-    basestring = str
+   from functools import cmp_to_key
 
 NoneType = type(None)
 try:
@@ -48,7 +47,8 @@ except:
 
 (F_LINK,
  F_LINKABLE,
- F_OPTIONAL) = [ 1<<x for x in range(3) ]
+ F_OPTIONAL,
+ F_SECRET) = [ 1<<x for x in range(4) ]
 
 BASE_VERSION = 'vmodl.version.version0'
 VERSION1     = 'vmodl.version.version1'
@@ -79,6 +79,9 @@ _wsdlDefMap = {}
 # if a.b.c and a.b.d are the nested classes of a.b, then _dependencyMap[a.b] = {c,d}
 _dependencyMap = {}
 
+# Map top level names to xml namespaces
+_urnMap = {"vim": XMLNS_VMODL_BASE, "sms": "urn:sms", "pbm": "urn:pbm"}
+
 ## Update the dependency map
 #  Note: Must be holding the _lazyLock
 #
@@ -191,7 +194,7 @@ class LazyObject(Object):
 
 class Link(text_type):
    def __new__(cls, obj):
-      if isinstance(obj, basestring):
+      if isinstance(obj, string_types):
          return text_type.__new__(cls, obj)
       elif isinstance(obj, DataObject):
          if obj.key:
@@ -251,9 +254,9 @@ class LazyModule(object):
          else:
             if _CheckForDependency(self.name, attr):
                typeObj = LazyModule(name)
-            elif self.name == "vim":
+            elif self.name in _urnMap:
                try:
-                  typeObj = GetWsdlType(XMLNS_VMODL_BASE, attr)
+                  typeObj = GetWsdlType(_urnMap[self.name], attr)
                except:
                   raise AttributeError(attr)
             else:
@@ -271,6 +274,131 @@ class LazyModule(object):
       else:
          raise AttributeError("'%s' does not exist" % self.name)
 
+
+class VmomiJSONEncoder(json.JSONEncoder):
+    """
+        Custom JSON encoder to encode vSphere objects.
+
+        When a ManagedObject is encoded, it gains three properties:
+            _vimid is the _moId (ex: 'vm-42')
+            _vimref is the moRef (ex: 'vim.VirtualMachine:vm-42')
+            _vimtype is the class name (ex: 'vim.VirtualMachine')
+
+        When a DataObject is encoded, it gains one property:
+            _vimtype is the class name (ex: 'vim.VirtualMachineQuestionInfo')
+
+        If the dynamicProperty and dynamicType are empty, they are optionally
+            omitted from the results of DataObjects and ManagedObjects
+
+        @example "Explode only the object passed in"
+            data = json.dumps(vm, cls=VmomiJSONEncoder)
+
+        @example "Explode specific objects"
+            data = json.dumps(vm, cls=VmomiJSONEncoder,
+                              explode=[vm, vm.network[0]])
+
+        @example "Explode all virtual machines in a list and their snapshots"
+            data = json.dumps([vm1, vm2], cls=VmomiJSONEncoder,
+                              explode=[templateOf('VirtualMachine'),
+                                       templateOf('VirtualMachineSnapshot')])
+    """
+    def __init__(self, *args, **kwargs):
+        """
+        Consumer must specify what types to explode into full content
+        and what types to leave as a ManagedObjectReference.  If the
+        root object of the encoding is a ManagedObject, it will be
+        added to the explode list automatically.
+
+        A ManagedObject is only exploded once, the first time it is
+        encountered.  All subsequent times it will be a moRef.
+
+        @param explode (list) A list of objects and/or types to
+            expand in the conversion process.  Directly add any
+            objects to the list, or add the type of any object
+            using type(templateOf('VirtualMachine')) for example.
+
+        @param strip_dynamic (bool) Every object normally contains
+            a dynamicProperty list and a dynamicType field even if
+            the contents are [] and None respectively.  These fields
+            clutter the output making it more difficult for a person
+            to read and bloat the size of the result unnecessarily.
+            This option removes them if they are the empty values above.
+            The default is False.
+        """
+        self._done = set()
+        self._first = True
+        self._explode = kwargs.pop('explode', None)
+        self._strip_dynamic = kwargs.pop('strip_dynamic', False)
+        super(VmomiJSONEncoder, self).__init__(*args, **kwargs)
+
+    def _remove_empty_dynamic_if(self, result):
+        if self._strip_dynamic:
+            if 'dynamicProperty' in result and len(result['dynamicProperty']) == 0:
+                result.pop('dynamicProperty') 
+            if 'dynamicType' in result and not result['dynamicType']:
+                result.pop('dynamicType')
+        return result
+
+    def default(self, obj):  # pylint: disable=method-hidden
+        if self._first:
+            self._first = False
+            if not self._explode:
+                self._explode = []
+            if isinstance(obj, ManagedObject):
+                self._explode.append(obj)
+        if isinstance(obj, ManagedObject):
+            if self.explode(obj):
+                result = {}
+                result['_vimid'] = obj._moId
+                result['_vimref'] = '{}:{}'.format(obj.__class__.__name__,
+                                                   obj._moId)
+                result['_vimtype'] = obj.__class__.__name__
+                for prop in obj._GetPropertyList():
+                    result[prop.name] = getattr(obj, prop.name)
+                return self._remove_empty_dynamic_if(result)
+            return str(obj).strip("'")  # see FormatObject below
+        if isinstance(obj, DataObject):
+            result = {k:v for k,v in obj.__dict__.items()}
+            result['_vimtype'] = obj.__class__.__name__
+            return self._remove_empty_dynamic_if(result)
+        if isinstance(obj, binary):
+            result = base64.b64encode(obj)
+            if PY3:  # see FormatObject below
+                result = str(result, 'utf-8')
+            return result
+        if isinstance(obj, datetime):
+            return Iso8601.ISO8601Format(obj)
+        if isinstance(obj, UncallableManagedMethod):
+            return obj.name
+        if isinstance(obj, ManagedMethod):
+            return str(obj)  # see FormatObject below
+        if isinstance(obj, type):
+            return obj.__name__
+        super(VmomiJSONEncoder, self).default(obj)
+
+    def explode(self, obj):
+        """ Determine if the object should be exploded. """
+        if obj in self._done:
+            return False
+        result = False
+        for item in self._explode:
+            if hasattr(item, '_moId'):
+                # If it has a _moId it is an instance
+                if obj._moId == item._moId:
+                    result = True
+            else:
+                # If it does not have a _moId it is a template
+                if obj.__class__.__name__ == item.__name__:
+                    result = True
+        if result:
+            self._done.add(obj)
+        return result
+
+
+def templateOf(typestr):
+    """ Returns a class template. """
+    return GetWsdlType(XMLNS_VMODL_BASE, typestr)
+
 ## Format a python VMOMI object
 #
 # @param val the object
@@ -305,6 +433,8 @@ def FormatObject(val, info=Object(name="", type=object, flags=0), indent=0):
          result = "(%s) []" % itemType.__name__
    elif isinstance(val, type):
       result = val.__name__
+   elif isinstance(val, UncallableManagedMethod):
+      result = val.name
    elif isinstance(val, ManagedMethod):
       result = '%s.%s' % (val.info.type.__name__, val.info.name)
    elif isinstance(val, bool):
@@ -313,6 +443,13 @@ def FormatObject(val, info=Object(name="", type=object, flags=0), indent=0):
       result = Iso8601.ISO8601Format(val)
    elif isinstance(val, binary):
       result = base64.b64encode(val)
+      if PY3:
+         # In python3 the bytes result after the base64 encoding has a
+         # leading 'b' which causes error when we use it to construct the
+         # soap message. Workaround the issue by converting the result to
+         # string. Since the result of base64 encoding contains only subset
+         # of ASCII chars, converting to string will not change the value.
+         result = str(result, 'utf-8')
    else:
       result = repr(val)
    return start + result
@@ -371,7 +508,7 @@ class ManagedObject(object):
                raise TypeError("%s() got multiple values for keyword argument '%s'" %
                               (Capitalize(info.name),  k))
             args[idx] = v
-      map(CheckField, info.params, args)
+      list(map(CheckField, info.params, args))
       return self._stub.InvokeMethod(self, info, args)
    _InvokeMethod = staticmethod(_InvokeMethod)
 
@@ -419,7 +556,7 @@ class ManagedObject(object):
       result = []
       while cls != ManagedObject:
          # Iterate through methods, add info for method not found in derived class
-         result = [info for info in cls._methodInfo.values()
+         result = [info for info in list(cls._methodInfo.values())
                    if meth.setdefault(info.name, cls) == cls] + result
          cls = cls.__bases__[0]
       return result
@@ -464,6 +601,9 @@ class ManagedObject(object):
                 self.__class__ == other.__class__ and \
                 self._serverGuid == other._serverGuid
 
+   def __ne__(self, other):
+      return not(self == other)
+
    def __hash__(self):
       return str(self).__hash__()
 
@@ -500,7 +640,7 @@ class DataObject(Base):
             SetAttr(self, info.name, info.type(0))
          else:
             SetAttr(self, info.name, None)
-      for (k, v) in kwargs.items():
+      for (k, v) in list(kwargs.items()):
          setattr(self, k, v)
 
    ## Get a list of all properties of this type and base types
@@ -576,6 +716,15 @@ class ManagedMethod(Curry):
       Curry.__init__(self, ManagedObject._InvokeMethod, info)
       self.info = info
 
+# Method used to represent any unknown wsdl method returned by server response.
+# Server may return unknown method name due to server defects or newer version.
+class UncallableManagedMethod(ManagedMethod):
+   def __init__(self, name):
+      self.name = name
+
+   def __call__(self, *args, **kwargs):
+      raise Exception("Managed method {} is not available".format(self.name))
+
 ## Create the vmodl.MethodFault type
 #
 # This type must be generated dynamically because it inherits from
@@ -939,7 +1088,7 @@ def CheckField(info, val):
         or issubclass(info.type, long) and (issubclass(valType, int) or \
                                             issubclass(valType, long)) \
         or issubclass(info.type, float) and issubclass(valType, float) \
-        or issubclass(info.type, basestring) and issubclass(valType, basestring):
+        or issubclass(info.type, string_types) and issubclass(valType, string_types):
          return
       elif issubclass(info.type, Link):
          # Allow object of expected type to be assigned to link
@@ -956,9 +1105,9 @@ def FinalizeType(type):
       for info in type._propList:
          info.type = GetVmodlType(info.type)
    elif issubclass(type, ManagedObject):
-      for info in type._propInfo.values():
+      for info in list(type._propInfo.values()):
          info.type = GetVmodlType(info.type)
-      for info in type._methodInfo.values():
+      for info in list(type._methodInfo.values()):
          info.result = GetVmodlType(info.result)
          info.methodResult = GetVmodlType(info.methodResult)
          info.type = GetVmodlType(info.type)
@@ -1026,9 +1175,15 @@ class UnknownWsdlTypeError(KeyError):
 # @return type if found in any one of the name spaces else throws KeyError
 def GuessWsdlType(name):
    with _lazyLock:
-      # Because the types are lazily loaded, if some name is present
-      # in multiple namespaces, we will load the first type that we
-      # encounter and return it.
+      # Some types may exist in multiple namespaces, and returning
+      # the wrong one will cause a deserialization error.
+      # Since in python3 the order of entries in set is not deterministic,
+      # we will try to get the type from vim25 namespace first.
+      try:
+         return GetWsdlType(XMLNS_VMODL_BASE, name)
+      except KeyError:
+         pass
+
       for ns in _wsdlTypeMapNSs:
          try:
             return GetWsdlType(ns, name)
@@ -1089,7 +1244,20 @@ def UncapitalizeVmodlName(str):
 
 ## Add a parent version
 def AddVersionParent(version, parent):
-   parentMap[version][parent] = True
+   parentMap[version].add(parent)
+
+def GetVersionProps(version):
+   """Get version properties
+
+   This function is a fixed version of GetVersion().
+   """
+
+   ns = nsMap[version]
+   versionId = versionIdMap[version]
+   isLegacy = versionMap.get(ns) == version
+   serviceNs =  serviceNsMap[version]
+   return ns, versionId, isLegacy, serviceNs
+
 
 ## Get version namespace from version
 def GetVersionNamespace(version):
@@ -1113,6 +1281,10 @@ def GetWsdlNamespace(version):
    """ Get wsdl namespace from version """
    return "urn:" + serviceNsMap[version]
 
+## Get an iterable with all version parents
+def GetVersionParents(version):
+   return parentMap[version]
+
 ## Get all the versions for the service with specified namespace (partially) ordered
 ## by compatibility (i.e. any version in the list that is compatible with some version
 ## v in the list will preceed v)
@@ -1133,22 +1305,21 @@ def GetServiceVersions(namespace):
    by compatibility (i.e. any version in the list that is compatible with some version
    v in the list will preceed v)
    """
-   versions = dict((v, True) for (v, n) in iteritems(serviceNsMap) if n == namespace)
-   mappings = {}
-   for v in iterkeys(versions):
-      mappings[v] = set(parent for parent in iterkeys(parentMap[v])
-                        if parent != v and parent in versions.keys())
-   res = []
-   while True:
-      el = [ k for (k, v) in iteritems(mappings) if len(v) == 0 ]
-      if len(el) == 0:
-         return res
-      el.sort()
-      for k in el:
-         res.insert(0, k)
-         del mappings[k]
-         for values in itervalues(mappings):
-            values.discard(k)
+   def compare(a, b):
+      if a == b:
+         return 0
+      if b in parentMap[a]:
+         return -1
+      if a in parentMap[b]:
+         return 1
+      return (a > b) - (a < b)
+
+   if PY3:
+      return sorted([v for (v, n) in iteritems(serviceNsMap) if n == namespace],
+                    key=cmp_to_key(compare))
+   else:
+      return sorted([v for (v, n) in iteritems(serviceNsMap) if n == namespace],
+                    compare)
 
 
 ## Set a WSDL method with wsdl namespace and wsdl name
@@ -1220,6 +1391,15 @@ def GetWsdlMethod(ns, wsdlName):
 # KeyError
 def GuessWsdlMethod(name):
    with _lazyLock:
+      # Some methods may exist in multiple namespaces, and returning
+      # the wrong one will cause a deserialization error.
+      # Since in python3 the order of entries in set is not deterministic,
+      # we will try to get the method from vim25 namespace first.
+      try:
+         return GetWsdlMethod(XMLNS_VMODL_BASE, name)
+      except KeyError:
+         pass
+
       for ns in _wsdlMethodNSs:
          try:
             return GetWsdlMethod(ns, name)
@@ -1240,6 +1420,11 @@ def GetCompatibleType(type, version):
 def InverseMap(map):
    return dict([ (v, k) for (k, v) in iteritems(map) ])
 
+def GetVmodlNs(version):
+   versionParts = version.split('.version.')
+   assert len(versionParts) == 2, 'Unsupported version format: %s' % version
+   return versionParts[0]
+
 types = Object()
 nsMap = {}
 versionIdMap = {}
@@ -1247,7 +1432,107 @@ versionMap = {}
 serviceNsMap = { BASE_VERSION : XMLNS_VMODL_BASE.split(":")[-1] }
 parentMap = {}
 
-from pyVmomi.Version import AddVersion, IsChildVersion
+class _MaturitySet:
+   """
+   Registry for versions from all namespaces defining a given maturity.
+   The registration is automatic (relevant code is generated by emitters),
+   while for the query one may use either the VMODL namespace id (e.g. 'vim'),
+   or the wire namespace id (e.g. 'vim25').
+   """
+   def __init__(self):
+      self._verNameMap = {}   # e.g. 'vim'   -> 'vim.version.version12'
+      self._verNameMapW = {}  # e.g. 'vim25' -> 'vim.version.version12'
+      self._wireIdMap = {}    # e.g. 'vim'   -> 'vim25/6.7'
+      self._wireIdMapW = {}   # e.g. 'vim25' -> 'vim25/6.7'
+
+   def Add(self, version):
+      """
+      Register the version at corresponding maturity for a given VMODL
+      namespace. The 'version' parameter is in the VMODL name format
+      e.g. 'vim.version.version12'. This method is typically used by
+      auto-generated code.
+      """
+      vmodlNs = GetVmodlNs(version)
+
+      # TODO fix the VSAN-related part of vcenter-all to enable the assert
+      # assert not (vmodlNs in self._verNameMap), 'Re-definition: %s' % vmodlNs
+
+      wireId = GetVersionNamespace(version)
+      wireNs = wireId.split('/')[0]
+      self._verNameMap[vmodlNs] = version
+      self._verNameMapW[wireNs] = version
+      self._wireIdMap[vmodlNs] = wireId
+      self._wireIdMapW[wireNs] = wireId
+      return wireId, wireNs
+
+   def GetName(self, vmodlNs):
+      """
+      VMODL namespace to registered version name mapping, e.g.
+      'vim' -> 'vim.version.version12'
+      """
+      return self._verNameMap[vmodlNs]
+
+   def GetNameW(self, wireNs):
+      """
+      Wire namespace to registered version name mapping, e.g.
+      'vim25' -> 'vim.version.version12'
+      """
+      return self._verNameMapW[wireNs]
+
+   def GetWireId(self, vmodlNs):
+      """
+      VMODL namespace to registered version wire-id mapping, e.g.
+      'vim' -> 'vim25/6.7'
+      """
+      return self._wireIdMap[vmodlNs]
+
+   def GetWireIdW(self, wireNs):
+      """
+      Wire namespace to registered version wire-id mapping, e.g.
+      'vim25' -> 'vim25/6.7'
+      """
+      return self._wireIdMapW[wireNs]
+
+   def EnumerateVmodlNs(self):
+      """
+      Returns an iterable with registered VMODL namespace, e.g.
+      ['vim', 'vpx', ... ]
+      """
+      return self._verNameMap.keys()
+
+   def EnumerateWireNs(self):
+      """
+      Returns an iterable with registered wire namespace, e.g.
+      ['vim25', 'vpxd3', ... ]
+      """
+      return self._verNameMapW.keys()
+
+   def EnumerateVersions(self):
+      """
+      Returns an iterable with registered VMODL versions, e.g.
+      ['vim.version.version12', 'vpx.version.version12', ... ]
+      """
+      return self._verNameMap.values()
+
+   def EnumerateWireIds(self):
+      """
+      Returns an iterable with registered versions wire-ids, e.g.
+      e.g. ['vim25/6.7', 'vpxd3/6.7', ... ]
+      """
+      return self._wireIdMap.values()
+
+# Backward compatibility aliases
+_MaturitySet.Get = _MaturitySet.GetName
+_MaturitySet.GetNamespace = _MaturitySet.GetWireId
+
+
+newestVersions = _MaturitySet()
+stableVersions = _MaturitySet()
+publicVersions = _MaturitySet()
+dottedVersions = _MaturitySet()
+oldestVersions = _MaturitySet()
+
+from .Version import AddVersion, IsChildVersion
 
 if not isinstance(bool, type): # bool not a type in python <= 2.2
    bool = type("bool", (int,),
@@ -1255,8 +1540,17 @@ if not isinstance(bool, type): # bool not a type in python <= 2.2
 byte  = type("byte", (int,), {})
 short  = type("short", (int,), {})
 double = type("double", (float,), {})
+if PY3:
+   long = type("long", (int,), {})
 URI = type("URI", (str,), {})
-binary = type("binary", (str,), {})
+if not PY3:
+    # six defines binary_type in python2 as a string; this means the
+    # JSON encoder sees checksum properties as strings and attempts
+    # to perform utf-8 decoding on them because they contain high-bit
+    # characters.
+    binary = type("binary", (bytearray,), {})
+else:
+    binary = type("binary", (binary_type,), {})
 PropertyPath = type("PropertyPath", (text_type,), {})
 
 # _wsdlTypeMapNSs store namespaces added to _wsdlTypeMap in _SetWsdlType
@@ -1283,7 +1577,7 @@ _wsdlTypeMap = {
 }
 _wsdlNameMap = InverseMap(_wsdlTypeMap)
 
-for ((ns, name), typ) in iteritems(dict(_wsdlTypeMap)):
+for ((ns, name), typ) in list(_wsdlTypeMap.items()):
    if typ is not NoneType:
       setattr(types, typ.__name__, typ)
       _wsdlTypeMapNSs.add(ns)
@@ -1487,7 +1781,7 @@ class StringDict(dict):
 
    def __setitem__(self, key, val):
       """x.__setitem__(i, y) <==> x[i]=y, where y must be a string"""
-      if not isinstance(val, basestring):
+      if not isinstance(val, string_types):
          raise TypeError("key %s has non-string value %s of %s" %
                                                          (key, val, type(val)))
       return dict.__setitem__(self, key, val)
@@ -1602,7 +1896,7 @@ class LinkResolver:
    def _AddLinkable(self, obj):
       key = getattr(obj, "key")
       if key and key != '':
-         if self.linkables.has_key(key):
+         if key in self.linkables:
             #duplicate key present
             raise AttributeError(key)
          else:
@@ -1648,3 +1942,17 @@ def ResolveLinks(keys, obj):
       return None
    linkResolver = LinkResolver(obj)
    return linkResolver.ResolveLinks(keys)
+
+
+# Dictionary of type { 'branch' : { 'namespace' : count } }
+# used by Vmodl Vcdb B2B code.
+
+_breakingChanges = {}
+
+
+def AddBreakingChangesInfo(branchName, vmodlNamespace, count):
+   _breakingChanges.setdefault(branchName, {})[vmodlNamespace] = count
+
+
+def GetBreakingChanges():
+   return _breakingChanges
